/********************************************************************************
** Form generated from reading UI file 'QtGnuplotSettings.ui'
**
** Created by: Qt User Interface Compiler version 5.9.4
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_QTGNUPLOTSETTINGS_H
#define UI_QTGNUPLOTSETTINGS_H

#include <QtCore/QVariant>
#include <QtWidgets/QAction>
#include <QtWidgets/QApplication>
#include <QtWidgets/QButtonGroup>
#include <QtWidgets/QCheckBox>
#include <QtWidgets/QComboBox>
#include <QtWidgets/QDialog>
#include <QtWidgets/QDialogButtonBox>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QHeaderView>
#include <QtWidgets/QLabel>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QSpacerItem>
#include <QtWidgets/QVBoxLayout>

QT_BEGIN_NAMESPACE

class Ui_settingsDialog
{
public:
    QVBoxLayout *verticalLayout_2;
    QVBoxLayout *verticalLayout;
    QHBoxLayout *horizontalLayout;
    QPushButton *backgroundButton;
    QSpacerItem *horizontalSpacer;
    QLabel *sampleColorLabel;
    QCheckBox *antialiasCheckBox;
    QCheckBox *replotOnResizeCheckBox;
    QCheckBox *roundedCheckBox;
    QCheckBox *ctrlQCheckBox;
    QHBoxLayout *horizontalLayout_2;
    QLabel *label;
    QComboBox *mouseLabelComboBox;
    QDialogButtonBox *buttonBox;

    void setupUi(QDialog *settingsDialog)
    {
        if (settingsDialog->objectName().isEmpty())
            settingsDialog->setObjectName(QStringLiteral("settingsDialog"));
        settingsDialog->resize(269, 186);
        QIcon icon;
        icon.addFile(QStringLiteral(":/images/settings"), QSize(), QIcon::Normal, QIcon::Off);
        settingsDialog->setWindowIcon(icon);
        verticalLayout_2 = new QVBoxLayout(settingsDialog);
        verticalLayout_2->setObjectName(QStringLiteral("verticalLayout_2"));
        verticalLayout = new QVBoxLayout();
        verticalLayout->setObjectName(QStringLiteral("verticalLayout"));
        horizontalLayout = new QHBoxLayout();
        horizontalLayout->setObjectName(QStringLiteral("horizontalLayout"));
        backgroundButton = new QPushButton(settingsDialog);
        backgroundButton->setObjectName(QStringLiteral("backgroundButton"));

        horizontalLayout->addWidget(backgroundButton);

        horizontalSpacer = new QSpacerItem(40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum);

        horizontalLayout->addItem(horizontalSpacer);

        sampleColorLabel = new QLabel(settingsDialog);
        sampleColorLabel->setObjectName(QStringLiteral("sampleColorLabel"));

        horizontalLayout->addWidget(sampleColorLabel);


        verticalLayout->addLayout(horizontalLayout);

        antialiasCheckBox = new QCheckBox(settingsDialog);
        antialiasCheckBox->setObjectName(QStringLiteral("antialiasCheckBox"));

        verticalLayout->addWidget(antialiasCheckBox);

        replotOnResizeCheckBox = new QCheckBox(settingsDialog);
        replotOnResizeCheckBox->setObjectName(QStringLiteral("replotOnResizeCheckBox"));

        verticalLayout->addWidget(replotOnResizeCheckBox);

        roundedCheckBox = new QCheckBox(settingsDialog);
        roundedCheckBox->setObjectName(QStringLiteral("roundedCheckBox"));

        verticalLayout->addWidget(roundedCheckBox);

        ctrlQCheckBox = new QCheckBox(settingsDialog);
        ctrlQCheckBox->setObjectName(QStringLiteral("ctrlQCheckBox"));

        verticalLayout->addWidget(ctrlQCheckBox);

        horizontalLayout_2 = new QHBoxLayout();
        horizontalLayout_2->setObjectName(QStringLiteral("horizontalLayout_2"));
        label = new QLabel(settingsDialog);
        label->setObjectName(QStringLiteral("label"));
        label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);

        horizontalLayout_2->addWidget(label);

        mouseLabelComboBox = new QComboBox(settingsDialog);
        mouseLabelComboBox->setObjectName(QStringLiteral("mouseLabelComboBox"));

        horizontalLayout_2->addWidget(mouseLabelComboBox);


        verticalLayout->addLayout(horizontalLayout_2);


        verticalLayout_2->addLayout(verticalLayout);

        buttonBox = new QDialogButtonBox(settingsDialog);
        buttonBox->setObjectName(QStringLiteral("buttonBox"));
        buttonBox->setOrientation(Qt::Horizontal);
        buttonBox->setStandardButtons(QDialogButtonBox::Cancel|QDialogButtonBox::Ok);

        verticalLayout_2->addWidget(buttonBox);


        retranslateUi(settingsDialog);
        QObject::connect(buttonBox, SIGNAL(accepted()), settingsDialog, SLOT(accept()));
        QObject::connect(buttonBox, SIGNAL(rejected()), settingsDialog, SLOT(reject()));

        QMetaObject::connectSlotsByName(settingsDialog);
    } // setupUi

    void retranslateUi(QDialog *settingsDialog)
    {
        settingsDialog->setWindowTitle(QApplication::translate("settingsDialog", "Terminal configuration", Q_NULLPTR));
        backgroundButton->setText(QApplication::translate("settingsDialog", "Select background color", Q_NULLPTR));
        sampleColorLabel->setText(QApplication::translate("settingsDialog", "Sample", Q_NULLPTR));
        antialiasCheckBox->setText(QApplication::translate("settingsDialog", "Antialias", Q_NULLPTR));
        replotOnResizeCheckBox->setText(QApplication::translate("settingsDialog", "Replot on resize", Q_NULLPTR));
        roundedCheckBox->setText(QApplication::translate("settingsDialog", "Rounded line ends", Q_NULLPTR));
        ctrlQCheckBox->setText(QApplication::translate("settingsDialog", "ctrl-q closes plot window", Q_NULLPTR));
        label->setText(QApplication::translate("settingsDialog", "Mouse label", Q_NULLPTR));
        mouseLabelComboBox->clear();
        mouseLabelComboBox->insertItems(0, QStringList()
         << QApplication::translate("settingsDialog", "Status bar", Q_NULLPTR)
         << QApplication::translate("settingsDialog", "Tool bar", Q_NULLPTR)
         << QApplication::translate("settingsDialog", "Above plot", Q_NULLPTR)
         << QApplication::translate("settingsDialog", "None", Q_NULLPTR)
        );
    } // retranslateUi

};

namespace Ui {
    class settingsDialog: public Ui_settingsDialog {};
} // namespace Ui

QT_END_NAMESPACE

#endif // UI_QTGNUPLOTSETTINGS_H
